﻿/* _
MIT License

Copyright (c) 2022 zgock.Blazedust MacGruber

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE. 
*/
using System;
using UnityEngine;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;
namespace PluginBuilder_Zgock {
    class CUAAutoFollow : MVRScript {
        // Security and initialization
        private bool editorInitialized = false;
        private int initLoops = 0;
        private int maxInitLoops = 3600;

        private Transform _root = null;
        private Transform _parent = null;
        private Atom _parentAtom = null;


        Vector3 linkPos;
        Vector3 linkRot;

        private DAZCharacterSelector _characterSelector;

        //public JSONStorableBool jbEnable;
        private UIDynamicButton btnSave;
        private UIDynamicButton btnLoad;

        public JSONStorableFloat jfOffsetX;
        public JSONStorableFloat jfOffsetY;
        public JSONStorableFloat jfOffsetZ;

        public JSONStorableFloat jfRotateX;
        public JSONStorableFloat jfRotateY;
        public JSONStorableFloat jfRotateZ;


        public JSONStorableFloat jfScaleX;
        public JSONStorableFloat jfScaleY;
        public JSONStorableFloat jfScaleZ;

        public JSONStorableString jsRootName;
        public JSONStorableString jsScaleBase;

        private JSONStorableStringChooser jssMorphKeys = null;
        private JSONStorableStringChooser jssBoneKeys = null;

        private UIDynamicButton btnAddFromActive = null;
        private UIDynamicButton btnAddFromFav = null;
        private UIDynamicButton btnDelete = null;
        private UIDynamicButton btnAutoCalc = null;

        private List<string> _morphKeys = new List<string>();

        private string _currentMorph = "None";

        private List<string> _currentBones = new List<string>();
        private string _currentBone = "None";


        private List<string> _cuaBones = new List<string>();

        private List<string> _dazBones = new List<string>();


        bool bonebinded;

        bool _dirty = true;

        class MoveInfo {
            public Vector3 pos;
            public Quaternion rot;
            public Vector3 scale;
        }

        class cuaBone {
            public DAZBone bone;
            public Transform link;
            public Vector3 linkPos;
            public Vector3 linkRot;
            public Vector3 holdPos;
            public Quaternion holdRot;
            public Transform src;
            public Transform target;
            public Rigidbody rb;
            public bool posLinked;
            public bool rotLinked;
            public Vector3 posOffset;
            public Vector3 rotOffset;
            public Vector3 scaleOffset;
        }

        private Transform _g2fbase = null;

        private List<cuaBone> cuaBones = new List<cuaBone>();

        Dictionary<string, cuaBone> cuaDic = new Dictionary<string, cuaBone>();

        class cuaBoneInfo {
            public Transform trns;
            public Vector3 initialPos;
            public Quaternion initialRot;
        }

        private Dictionary<string, cuaBoneInfo> cuaBonedict = new Dictionary<string, cuaBoneInfo>();

        private List<JSONStorableFloat> cuaMorphs = new List<JSONStorableFloat>();

        private Dictionary<string, float> _blend = new Dictionary<string, float>();

        private Dictionary<string, float> _rate = new Dictionary<string, float>();
        private Dictionary<string, float> _lastrate = new Dictionary<string, float>();

        class MorphInfo {
            public DAZMorph morph;
            public float value;
            public float lastvalue;
        }

        private Dictionary<string, List<MorphInfo>> _morphs = new Dictionary<string, List<MorphInfo>>();

        public bool posLinked = false;
        public bool rotLinked = false;
        private bool trnLinked = false;

        private string saveFolder;
        private string assetFolder;
        private string defaultFolder;
        private string saveName;

        class BoneOffset {
            public float offsetX;
            public float offsetY;
            public float offsetZ;
            public float rotateX;
            public float rotateY;
            public float rotateZ;
            public float scaleX;
            public float scaleY;
            public float scaleZ;
            public Dictionary<string, float> morphs;
        }

        private Dictionary<string, Dictionary<string, BoneOffset>> morphDic = new Dictionary<string, Dictionary<string, BoneOffset>>();

        private int ratecount = 0;
        private bool _loading = false;


        JSONStorableFloat jsScale;



        private Dictionary<string, string> _linkDic = new Dictionary<string, string>()
        {
            //{"hipControl","hip" },
            {"hipLink","hip" },
            //{"pelvisControl","pelvis" },
            {"pelvisLink","pelvis" },
            //{"chestControl","chest" },
            {"chestLink","chest" },
            //{"headControl","head" },
            {"headLink","head" },
            //{"lHandControl","lHand" },
            {"lHandLink","lHand" },
            //{"rHandControl","rHand" },
            {"rHandLink","rHand" },
            //{"lFootControl","lFoot" },
            {"lFootLink","lFoot" },
            //{"rFootControl","rFoot" },
            {"rFootLink","rFoot" },
            //{"neckControl","neck" },
            //{"rNippleControl","rNipple" },
            //{"lNippleControl","lNipple" },
            //{"lElbowControl","lForeArm" },
            {"lElbowLink","lForeArm" },
            //{"rElbowControl","rForeArm" },
            {"rElbowLink","rForeArm" },
            //{"lKneeControl","lShin" },
            {"lKneeLink","lShin" },
            //{"rKneeControl","rShin" },
            {"rKneeLink","rShin" },
            {"lToeControl","lToe" },
            {"rToeControl","rToe" },
            //{"abdomenControl","abdomen" },
            {"abdomenLink","abdomen" },
            //{"abdomen2Control","abdomen2" },
            {"abdomen2Link","abdomen2" },
            //{"lArmControl","lShldr" },
            //{"rArmControl","rShldr" },
            //{"lShoulderControl","lCollar" },
            //{"rShoulderControl","rCollar" },
        };

        public override void Init() {
            try {
                // We only authorize the script to be used by a CUA
                if (containingAtom.type != "CustomUnityAsset") {
                    logError("MyCUAEditor - Please add this script on a Custom Unity Asset ( script disabled ).");
                    return;
                }

                VerifyDic();

                LoadDic();

                _currentBones.Clear();
                _currentBones.Add("None");
                _currentBone = _currentBones[0];

                btnSave = CreateButton("Save", true);
                btnSave.button.onClick.AddListener(() => {
                    try {
                        SaveDic();
                    } catch (Exception e) {
                        logError(e.ToString());
                    }
                });

                btnLoad = CreateButton("Load", true);
                btnLoad.button.onClick.AddListener(() => {
                    try {
                        LoadDic();
                    } catch (Exception e) {
                        logError(e.ToString());
                    }
                });
                UIDynamicSlider slider;
                jfOffsetX = new JSONStorableFloat("X Offset", 0.0f, OnChangeOffsetX, -1f, 1f, false);
                slider = CreateSlider(jfOffsetX, true);
                slider.ConfigureQuickButtons(-0.001f, -0.01f, -0.1f, -1f, 0.001f, 0.01f, 0.1f, 1f);
                slider.valueFormat = "{0:.000}";

                jfOffsetY = new JSONStorableFloat("Y Offset", 0.0f, OnChangeOffsetY, -1f, 1f, false);
                slider = CreateSlider(jfOffsetY, true);
                slider.ConfigureQuickButtons(-0.001f, -0.01f, -0.1f, -1f, 0.001f, 0.01f, 0.1f, 1f);
                slider.valueFormat = "{0:.000}";

                jfOffsetZ = new JSONStorableFloat("Z Offset", 0.0f, OnChangeOffsetZ, -1f, 1f, false);
                slider = CreateSlider(jfOffsetZ, true);
                slider.ConfigureQuickButtons(-0.001f, -0.01f, -0.1f, -1f, 0.001f, 0.01f, 0.1f, 1f);
                slider.valueFormat = "{0:.000}";

                jfRotateX = new JSONStorableFloat("X Rotate", 0.0f, OnChangeRotateX, -180.0f, 180.0f);
                slider = CreateSlider(jfRotateX, true);
                slider.ConfigureQuickButtons(-0.001f, -0.01f, -0.1f, -1f, 0.001f, 0.01f, 0.1f, 1f);
                slider.valueFormat = "{0:.000}";

                jfRotateY = new JSONStorableFloat("Y Rotate", 0.0f, OnChangeRotateY, -180.0f, 180.0f);
                slider = CreateSlider(jfRotateY, true);
                slider.ConfigureQuickButtons(-0.001f, -0.01f, -0.1f, -1f, 0.001f, 0.01f, 0.1f, 1f);
                slider.valueFormat = "{0:.000}";

                jfRotateZ = new JSONStorableFloat("Z Rotate", 0.0f, OnChangeRotateZ, -180.0f, 180.0f);
                slider = CreateSlider(jfRotateZ, true);
                slider.ConfigureQuickButtons(-0.001f, -0.01f, -0.1f, -1f, 0.001f, 0.01f, 0.1f, 1f);
                slider.valueFormat = "{0:.000}";

                jfScaleX = new JSONStorableFloat("X Scale", 1.0f, OnChangeScaleX, 0, 2.0f, false);
                slider = CreateSlider(jfScaleX, true);
                slider.ConfigureQuickButtons(-0.001f, -0.01f, -0.1f, -1f, 0.001f, 0.01f, 0.1f, 1f);
                slider.valueFormat = "{0:.000}";

                jfScaleY = new JSONStorableFloat("Y Scale", 1.0f, OnChangeScaleY, 0, 2.0f, false);
                slider = CreateSlider(jfScaleY, true);
                slider.ConfigureQuickButtons(-0.001f, -0.01f, -0.1f, -1f, 0.001f, 0.01f, 0.1f, 1f);
                slider.valueFormat = "{0:.000}";

                jfScaleZ = new JSONStorableFloat("Z Scale", 1.0f, OnChangeScaleZ, 0, 2.0f, false);
                slider = CreateSlider(jfScaleZ, true);
                slider.ConfigureQuickButtons(-0.001f, -0.01f, -0.1f, -1f, 0.001f, 0.01f, 0.1f, 1f);
                slider.valueFormat = "{0:.000}";

                RefreshUI();
            } catch (Exception e) {
                SuperController.LogError("CUAAutoFollow - Exception caught: " + e);
            }
        }

        private string PathToFolder(string path) {
            string folder = "";
            if (path.Contains("/") || path.Contains("\\")) { folder = path.Substring(0, path.LastIndexOfAny(new char[] { '/', '\\' })); }
            if (folder.StartsWith(".")) { folder = folder.Substring(1); }
            if (folder.StartsWith("/")) { folder = folder.Substring(1); }
            int varPackageIndex = folder.IndexOf("/");
            if (varPackageIndex > 0 && folder.IndexOf(":/") == (varPackageIndex - 1)) {
                folder = folder.Substring(varPackageIndex + 1);
            }
            return folder;

        }
        private void VerifyDic() {
            string pluginId = this.storeId.Split('_')[0];
            MVRPluginManager manager = containingAtom.GetStorableByID("PluginManager") as MVRPluginManager;
            string pathToScriptFile = manager.GetJSON(true, true)["plugins"][pluginId].Value;
            logDebug($"Script file = {pathToScriptFile}");
            defaultFolder = PathToFolder(pathToScriptFile);

            JSONStorable js = containingAtom.GetStorableByID("asset");
            JSONStorableUrl jsu = js.GetUrlJSONParam("assetUrl");

            if (jsu != null) {
                logDebug($"Asset Path = {jsu.val}");
                int index = jsu.val.LastIndexOfAny(new char[] { '/', '\\', '/' });
                if (index > 0) {
                    saveName = jsu.val.Substring(index + 1);
                    if (saveName.IndexOf(".assetbundle") > 0) {
                        saveName = saveName.Substring(0, saveName.IndexOf(".assetbundle"));
                    } else {
                        logError("WARNING: assetbundle extention should be \".assetbundle\". Maybe it makes probrem in this plugin");
                    }
                }

            }
            assetFolder = PathToFolder(jsu.val);

            JSONStorableStringChooser jssc = js.GetStringChooserJSONParam("assetName");
            if (jssc != null) {
                logDebug($"Prehap Path  {jssc.val}");
                int index = jssc.val.LastIndexOfAny(new char[] { '/', '\\', '/' });
                if (index > 0) {
                    string s = jssc.val.Substring(index + 1);
                    if (s.IndexOf(".prefab") > 0) {
                        saveName += "_" + s.Substring(0, s.IndexOf(".prefab"));
                    } else if (s.IndexOf(".unity") > 0) {
                        saveName += "_" + s.Substring(0, s.IndexOf(".unity"));
                    } else {
                        saveName += "_" + s;
                        logError("WARNING: prefav extention should be \".prefab\" or \".unity\". Maybe it makes probrem in this plugin");
                        return;
                    }
                }

            }

            saveName += ".json";

            //saveFolder = "Saves/PluginData" + defaultFolder.Substring(defaultFolder.IndexOf("/Scripts/") + "/Scripts".Length);
            saveFolder = "Saves/PluginData/" + assetFolder.Substring(assetFolder.IndexOf("/Assets/") + "/Assets/".Length);
            logDebug($"Default dir = {defaultFolder}");
            logDebug($"Asset dir = {assetFolder}");
            logDebug($"Save dir = {saveFolder}");
            logDebug($"Save Name = {saveName}");
        }

        private void AddMorphKey(DAZMorph morph) {
            if (!_morphs.ContainsKey(morph.displayName)) {
                _morphs[morph.displayName] = new List<MorphInfo>();
            }
            bool replace = false;
            foreach (MorphInfo inf in _morphs[morph.displayName]) {
                if (inf.morph.GetHashCode() == morph.GetHashCode()) {
                    replace = true;
                    break;
                }
            }
            if (!replace) {
                MorphInfo info = new MorphInfo();
                info.morph = morph;
                _morphs[morph.displayName].Add(info);
            }
        }

        private void RebuildMorphs() {
            _morphs.Clear();
            if (_characterSelector != null) {
                foreach (KeyValuePair<string, Dictionary<string, BoneOffset>> pair in morphDic) {
                    _morphs[pair.Key] = new List<MorphInfo>();
                    foreach(DAZMorph morph in _characterSelector.morphBank1.morphs) {
                        if (morph.displayName == pair.Key) {
                            AddMorphKey(morph);
                        }
                    }
                }
            }
        }

        private void LoadDic() {
            _loading = true;
            bool saveExist = false, assetExist = false, defaultExist = false;
            try {
                string[] savePathes = SuperController.singleton.GetFilesAtPath(saveFolder, saveName);
                if (savePathes.Length > 0) {
                    saveExist = true;
                }
            } catch (Exception e) {
                saveExist = false;
            }
            try {
                string[] savePathes = SuperController.singleton.GetFilesAtPath(assetFolder, saveName);
                if (savePathes.Length > 0) {
                    assetExist = true;
                }
            } catch (Exception e) {
                assetExist = false;
            }
            try {
                string[] savePathes = SuperController.singleton.GetFilesAtPath(defaultFolder, saveName);
                if (savePathes.Length > 0) {
                    defaultExist = true;
                }
            } catch (Exception e) {
                defaultExist = false;
            }
            if (saveExist) {
                logDebug("savePath Exists");
                LoadDicFrom(saveFolder + "/" + saveName);
                logDebug("savePath Loaded");
            } else if (assetExist) {
                logDebug("assetPath Exists");
                LoadDicFrom(assetFolder + "/" + saveName);
                logDebug("assetPath Loaded");
            } else if (defaultExist) {
                logDebug("defaultPath Exists");
                LoadDicFrom(defaultFolder + "/" + saveName);
                logDebug("defaultPath Loaded");
            } else {
                logDebug("no Save File");
            }
            RebuildMorphs();
            RefreshUI();
            _loading = false;
        }

        private void LoadDicFrom(string path) {
            logDebug($"JSON Loading from {path}");
            JSONNode node = SuperController.singleton.LoadJSON(path);
            JSONClass jc = node.AsObject;
            morphDic.Clear();
            if (node["morphs"] == null) {
                logDebug("morphs is null");
                return;
            } else {
                logDebug(node["morphs"][0]["name"]);
                logDebug(node["morphs"][0]["bones"][0]["name"]);
                logDebug($"{node["morphs"][0]["bones"][0]["offsetX"].AsFloat}");
                if (node["morphs"][0]["bones"][0]["morphs"].AsArray.Count > 0) {
                    logDebug($"{node["morphs"][0]["bones"][0]["morphs"][0]["name"]}");
                    logDebug($"{node["morphs"][0]["bones"][0]["morphs"][0]["value"].AsFloat}");
                }
                for (int i = 0; i < node["morphs"].AsArray.Count; i++) {
                    Dictionary<string, BoneOffset> dic = new Dictionary<string, BoneOffset>();
                    for (int j = 0; j < node["morphs"][i]["bones"].AsArray.Count; j++) {
                        BoneOffset ofs = new BoneOffset();
                        ofs.offsetX = node["morphs"][i]["bones"][j]["offsetX"].AsFloat;
                        ofs.offsetY = node["morphs"][i]["bones"][j]["offsetY"].AsFloat;
                        ofs.offsetZ = node["morphs"][i]["bones"][j]["offsetZ"].AsFloat;
                        ofs.rotateX = node["morphs"][i]["bones"][j]["rotateX"].AsFloat;
                        ofs.rotateY = node["morphs"][i]["bones"][j]["rotateY"].AsFloat;
                        ofs.rotateZ = node["morphs"][i]["bones"][j]["rotateZ"].AsFloat;
                        ofs.scaleX = node["morphs"][i]["bones"][j]["scaleX"].AsFloat;
                        ofs.scaleY = node["morphs"][i]["bones"][j]["scaleY"].AsFloat;
                        ofs.scaleZ = node["morphs"][i]["bones"][j]["scaleZ"].AsFloat;
                        Dictionary<string, float> morph = new Dictionary<string, float>();
                        if (node["morphs"][i]["bones"][j]["morphs"] != null) {
                            for (int k = 0; k < node["morphs"][i]["bones"][j]["morphs"].AsArray.Count; k++) {
                                morph[node["morphs"][i]["bones"][j]["morphs"][k]["name"]] = node["morphs"][i]["bones"][j]["morphs"][k]["value"].AsFloat;
                            }
                        }
                        ofs.morphs = morph;
                        dic[node["morphs"][i]["bones"][j]["name"]] = ofs;
                    }
                    morphDic[node["morphs"][i]["name"]] = dic;
                }
            }
            _dirty = true;
            logDebug("JSON Loaded");
        }

        private void SaveDic() {
            logDebug($"Saving to {saveFolder}");
            JSONNode node = JSONNode.Parse("{}");

            foreach (KeyValuePair<string, Dictionary<string, BoneOffset>> pair in morphDic) {
                JSONNode morphs = JSONNode.Parse("{}");
                morphs["name"] = pair.Key;
                foreach (KeyValuePair<string, BoneOffset> pair2 in pair.Value) {
                    JSONNode offset = JSONNode.Parse("{}");
                    offset["name"] = pair2.Key;
                    offset["offsetX"] = pair2.Value.offsetX.ToString();
                    offset["offsetY"] = pair2.Value.offsetY.ToString();
                    offset["offsetZ"] = pair2.Value.offsetZ.ToString();
                    offset["rotateX"] = pair2.Value.rotateX.ToString();
                    offset["rotateY"] = pair2.Value.rotateY.ToString();
                    offset["rotateZ"] = pair2.Value.rotateZ.ToString();
                    offset["scaleX"] = pair2.Value.scaleX.ToString();
                    offset["scaleY"] = pair2.Value.scaleY.ToString();
                    offset["scaleZ"] = pair2.Value.scaleZ.ToString();
                    foreach (KeyValuePair<string, float> pair3 in pair2.Value.morphs) {
                        JSONNode morph = JSONNode.Parse("{}");
                        morph["name"] = pair3.Key;
                        morph["value"] = pair3.Value.ToString();
                        offset["morphs"].Add(morph);
                    }
                    morphs["bones"].Add(offset);
                }
                node["morphs"].Add(morphs);
            }
            JSONClass jc = node.AsObject;
            VerifyFolder();
            SuperController.singleton.SaveJSON(jc, saveFolder + "/" + saveName);
            logDebug($"Saved to {saveFolder}/{saveName}");
        }

        public bool VerifyFolder() {
            string[] existingFiles;
            try {
                existingFiles = SuperController.singleton.GetFilesAtPath(saveFolder, "*.json");
            } catch (Exception e) {
                SuperController.singleton.ClearErrors();
                if (e.GetType().FullName == "System.IO.DirectoryNotFoundException" || e.Message != null && e.Message.Contains("non-existent path")) {
                    logMessage("Missing folder '" + saveFolder + "'. Attempting to automatically create it.");
                    if (TryToCreateFolder()) {
                        return true;
                    }
                }
                logError("Looks like the folder '" + saveFolder + "' is missing. Plugin can't start without it. Create the folder and reload the plugin.");
                return false;
            }
            return true;
        }

        public bool TryToCreateFolder() {
            // Attempt to create the folder. Do this by saving a dummy-file
            Camera cam = SuperController.singleton.screenshotCamera;
            string nPath = saveFolder + "/" + saveName;
            try {

                // For the dummy-save to automatically create the directory, the path requires backslashes! A little quirk.
                nPath = nPath.Replace("/", "\\");

                // disable screenshot taking 
                // don't do this if you don't know what you're doing! 
                // This might not be allowed in future versions of VaM
                // we do restore it in the end of this function though!
                SuperController.singleton.screenshotCamera = null;

                // Save a dummy file only containing one atom to keep it small and speedy.
                Atom prioAtom = containingAtom;

                SuperController.singleton.Save(nPath, prioAtom, includePhysical: false, includeAppearance: false);

                // check files again!
                try {
                    string[] existingFiles = SuperController.singleton.GetFilesAtPath(saveFolder, "*.json");

                    // Can read files now, assuming folder was created!
                    SuperController.LogMessage("Creating folder successful!");

                    // plugin folder created!
                    return true;
                } catch (Exception exce) {
                    // Still no luck!
                }
            } catch (Exception ex) {
                // Even failed to create the directory with a dummy file... User must create this folder!
                SuperController.LogMessage("Automatic creation failed");
                SuperController.LogMessage(ex.ToString());
            } finally {
                SuperController.singleton.screenshotCamera = cam;
            }

            return false;
        }

        private void OnChangeOffsetX(float val) {
            if (_currentMorph != "None" && _currentBone != "None") {
                (morphDic[_currentMorph])[_currentBone].offsetX = val;
                _dirty = true;
            }
        }

        private void OnChangeOffsetY(float val) {
            if (_currentMorph != "None" && _currentBone != "None") {
                (morphDic[_currentMorph])[_currentBone].offsetY = val;
                _dirty = true;
            }
        }

        private void OnChangeOffsetZ(float val) {
            if (_currentMorph != "None" && _currentBone != "None") {
                (morphDic[_currentMorph])[_currentBone].offsetZ = val;
                logDebug($"new value = {(morphDic[_currentMorph])[_currentBone].offsetZ}");
                _dirty = true;
            }
        }

        private void OnChangeRotateX(float val) {
            if (_currentMorph != "None" && _currentBone != "None") {
                (morphDic[_currentMorph])[_currentBone].rotateX = val;
                _dirty = true;
            }
        }

        private void OnChangeRotateY(float val) {
            if (_currentMorph != "None" && _currentBone != "None") {
                (morphDic[_currentMorph])[_currentBone].rotateY = val;
                _dirty = true;
            }
        }

        private void OnChangeRotateZ(float val) {
            if (_currentMorph != "None" && _currentBone != "None") {
                (morphDic[_currentMorph])[_currentBone].rotateZ = val;
                _dirty = true;
            }
        }

        private void OnChangeScaleX(float val) {
            if (_currentMorph != "None" && _currentBone != "None") {
                (morphDic[_currentMorph])[_currentBone].scaleX = val;
                _dirty = true;
            }
        }

        private void OnChangeScaleY(float val) {
            if (_currentMorph != "None" && _currentBone != "None") {
                (morphDic[_currentMorph])[_currentBone].scaleY = val;
                _dirty = true;
            }
        }

        private void OnChangeScaleZ(float val) {
            if (_currentMorph != "None" && _currentBone != "None") {
                (morphDic[_currentMorph])[_currentBone].scaleZ = val;
                _dirty = true;
            }
        }

        private void OnChangeMorph(JSONStorableFloat jf) {
            if (_currentMorph != "None" && _currentBone != "None") {
                (morphDic[_currentMorph])[_currentBone].morphs[jf.name] = jf.val;
                _dirty = true;
            }
        }

        private void RefreshValues() {
            if (_currentMorph != "None" && _currentBone != "None") {
                if (!morphDic[_currentMorph].ContainsKey(_currentBone)) {
                    BoneOffset ofs = new BoneOffset();
                    ofs.offsetX = 0;
                    ofs.offsetY = 0;
                    ofs.offsetZ = 0;
                    ofs.rotateX = 0;
                    ofs.rotateY = 0;
                    ofs.rotateZ = 0;
                    ofs.scaleX = 1;
                    ofs.scaleY = 1;
                    ofs.scaleZ = 1;
                    ofs.morphs = new Dictionary<string, float>();
                    foreach (JSONStorableFloat jf in cuaMorphs) {
                        ofs.morphs[jf.name] = 0.0f;
                    }
                    (morphDic[_currentMorph])[_currentBone] = ofs;
                    logDebug($"new bone entry{_currentBone} created");
                } else {
                    logDebug($"bone entry {_currentBone} already exists");
                }
                jfOffsetX.SetVal((morphDic[_currentMorph])[_currentBone].offsetX);
                jfOffsetY.SetVal((morphDic[_currentMorph])[_currentBone].offsetY);
                jfOffsetZ.SetVal((morphDic[_currentMorph])[_currentBone].offsetZ);

                jfRotateX.SetVal((morphDic[_currentMorph])[_currentBone].rotateX);
                jfRotateY.SetVal((morphDic[_currentMorph])[_currentBone].rotateY);
                jfRotateZ.SetVal((morphDic[_currentMorph])[_currentBone].rotateZ);

                jfScaleX.SetVal((morphDic[_currentMorph])[_currentBone].scaleX);
                jfScaleY.SetVal((morphDic[_currentMorph])[_currentBone].scaleY);
                jfScaleZ.SetVal((morphDic[_currentMorph])[_currentBone].scaleZ);

                foreach (JSONStorableFloat jf in cuaMorphs) {
                    jf.SetVal((morphDic[_currentMorph])[_currentBone].morphs[jf.name]);
                }
            }
        }

        private void RefreshUI() {
            string lastmorph = _currentMorph;
            if (jssBoneKeys != null) RemovePopup(jssBoneKeys);
            if (btnDelete != null) RemoveButton(btnDelete);
            if (btnAddFromFav != null) RemoveButton(btnAddFromFav);
            if (btnAddFromActive != null) RemoveButton(btnAddFromActive);
            if (jssMorphKeys != null) RemovePopup(jssMorphKeys);
            _morphKeys.Clear();
            foreach (KeyValuePair<string, Dictionary<string, BoneOffset>> pair in morphDic) {
                _morphKeys.Add(pair.Key);
            }
            if (_morphKeys.Count <= 0) {
                _morphKeys.Add("None");
            }
            if (!morphDic.ContainsKey(_currentMorph)) {
                _currentMorph = _morphKeys[0];
            }

            jssMorphKeys = new JSONStorableStringChooser("morphKeys", _morphKeys, _currentMorph, "Target", OnSelectMorph);
            UIDynamicPopup udp;
            udp = CreatePopup(jssMorphKeys);
            udp.labelWidth = 100f;
            btnAddFromActive = CreateButton("Add from Active");
            btnAddFromActive.button.onClick.AddListener(AddFromActive);
            btnAddFromFav = CreateButton("Add from Fav");
            btnAddFromFav.button.onClick.AddListener(AddFromFav);
            btnDelete = CreateButton("Delete");
            btnDelete.button.onClick.AddListener(DeleteCurrent);
            jssBoneKeys = new JSONStorableStringChooser("boneKeys", _currentBones, _currentBone, "Bone", OnSelectBone);
            udp = CreateScrollablePopup(jssBoneKeys);
            udp.labelWidth = 100f;

            RefreshValues();
            logDebug("UI Created");
        }

        private void OnSelectMorph(string val) {
            _currentMorph = val;
            RefreshValues();
            logDebug($"morph selected tp {_currentMorph}");
        }

        private void OnSelectBone(string val) {
            _currentBone = val;
            RefreshValues();
            logDebug($"bone selected tp {_currentBone}");
        }

        private void CreateMorphDic(DAZMorph morph) {
            if (!morphDic.ContainsKey(morph.displayName)) {
                Dictionary<string, BoneOffset> dic = new Dictionary<string, BoneOffset>();
                foreach (string name in _dazBones) {
                    BoneOffset ofs = new BoneOffset();
                    ofs.offsetX = 0;
                    ofs.offsetY = 0;
                    ofs.offsetZ = 0;
                    ofs.rotateX = 0;
                    ofs.rotateY = 0;
                    ofs.rotateZ = 0;
                    ofs.scaleX = 1;
                    ofs.scaleY = 1;
                    ofs.scaleZ = 1;
                    ofs.morphs = new Dictionary<string, float>();

                    foreach (JSONStorableFloat jf in cuaMorphs) {
                        ofs.morphs[jf.name] = 0.0f;
                    }
                    dic[name] = ofs;
                }
                foreach (cuaBone b in cuaBones) {
                    if (b.bone != null) continue;
                    BoneOffset ofs = new BoneOffset();
                    ofs.offsetX = 0;
                    ofs.offsetY = 0;
                    ofs.offsetZ = 0;
                    ofs.rotateX = 0;
                    ofs.rotateY = 0;
                    ofs.rotateZ = 0;
                    ofs.scaleX = 1;
                    ofs.scaleY = 1;
                    ofs.scaleZ = 1;
                    ofs.morphs = new Dictionary<string, float>();
                    foreach (JSONStorableFloat jf in cuaMorphs) {
                        ofs.morphs[jf.name] = 0.0f;
                    }
                    dic[b.target.name] = ofs;
                }
                morphDic[morph.displayName] = dic;

                logDebug($"new Morph {morph.displayName} add.");
            } else {
                var dic = morphDic[morph.displayName];
                foreach (cuaBone b in cuaBones) {
                    if (b.bone != null) continue;
                    if (dic.ContainsKey(b.target.name)) continue;
                    BoneOffset ofs = new BoneOffset();
                    ofs.offsetX = 0;
                    ofs.offsetY = 0;
                    ofs.offsetZ = 0;
                    ofs.rotateX = 0;
                    ofs.rotateY = 0;
                    ofs.rotateZ = 0;
                    ofs.scaleX = 1;
                    ofs.scaleY = 1;
                    ofs.scaleZ = 1;
                    ofs.morphs = new Dictionary<string, float>();
                    foreach (JSONStorableFloat jf in cuaMorphs) {
                        ofs.morphs[jf.name] = 0.0f;
                    }
                    dic[b.target.name] = ofs;
                }

                foreach (KeyValuePair<string, BoneOffset> ofs in dic) {
                    foreach (JSONStorableFloat jf in cuaMorphs) {
                        if (!ofs.Value.morphs.ContainsKey(jf.name)) {
                            ofs.Value.morphs[jf.name] = 0;
                            logDebug($"new Morph {jf.name} add.");
                        }
                    }
                }
            }
        }

        private void AddFromActive() {
            if (_characterSelector == null) return;
            foreach (DAZMorph morph in _characterSelector.morphBank1.morphs) {
                if (morph.active && morph.visible) {
                    AddMorphKey(morph);
                    CreateMorphDic(morph);
                }
            }
            RefreshUI();
            logDebug("Add From Active");
        }

        private void AddFromFav() {
            if (_characterSelector == null) return;
            foreach (DAZMorph morph in _characterSelector.morphBank1.morphs) {
                if (morph.favorite && morph.visible) {
                    AddMorphKey(morph);
                    CreateMorphDic(morph);
                }
            }
            RefreshUI();
            logDebug("Add From Fav");
        }

        private void DeleteCurrent() {
            if (_characterSelector == null) return;
            if (_currentMorph == "None") return;
            _morphs.Remove(_currentMorph);
            morphDic.Remove(_currentMorph);
            RebuildMorphs();
            RefreshUI();
        }

        void Update() {
            if (initLoops < maxInitLoops && editorInitialized == false) {
                initLoops++;
                InitEditor();
            }

            if (initLoops == maxInitLoops) {
                logError("CUAAutoFollow : This CUA could not be initialized properly after 3600 tries. Please select a compatible object and reload the script.");
                initLoops++; // To avoid spamming the log with the error
            }
            if (_root != null && editorInitialized) {
                //SyncWithBone();
                if (containingAtom.parentAtom != _parentAtom) {
                    _parentAtom = containingAtom.parentAtom;
                    _characterSelector = null;
                    for (Atom __parent = _parentAtom; __parent != null; __parent = __parent.parentAtom) {
                        if (__parent.type == "Person") {
                            DAZCharacterSelector c = __parent.GetComponentInChildren<DAZCharacterSelector>();
                            if (c != null) {
                                logDebug($"CUAAutoFollow : DAZCharacterSelector Found on {__parent.name}");
                                _characterSelector = c;
                                _parentAtom = __parent;

                                cuaBones.Clear();
                                cuaDic.Clear();

                                _dazBones.Clear();
                                _currentBones.Clear();



                                jsScale = _parentAtom.GetStorableByID("rescaleObject").GetFloatJSONParam("scale");
                                if (jsScale != null) {
                                    logDebug($"scale = {jsScale.val}");
                                } else {
                                    logDebug("scale not found");
                                }


                                foreach (DAZBone b in c.rootBones.dazBones) {
                                    _dazBones.Add(b.name);
                                    if (cuaBonedict.ContainsKey(b.name)) {
                                        _currentBones.Add(b.name);
                                        cuaBone pp = new cuaBone();
                                        pp.bone = b;
                                        pp.src = b.transform;
                                        pp.target = cuaBonedict[b.name].trns;
                                        pp.link = null;
                                        pp.posLinked = false;
                                        pp.rotLinked = false;
                                        cuaBones.Add(pp);
                                        cuaDic[b.name] = pp;
                                        logDebug($"bone binded to {b.name}");
                                    }
                                }
                                /*
                                foreach(KeyValuePair<string,cuaBoneInfo> info in cuaBonedict)
                                {
                                    Collider col = info.Value.trns.GetComponent<Collider>();
                                    if (col != null) continue;
                                    if (cuaDic.ContainsKey(info.Value.trns.parent.name) && !cuaDic.ContainsKey(info.Value.trns.name))
                                    {
                                        if (cuaDic[info.Value.trns.parent.name].bone == null) continue;
                                        _currentBones.Add(info.Value.trns.name);
                                        cuaBone pp = new cuaBone();
                                        pp.bone = null;
                                        pp.target = info.Value.trns;
                                        pp.linkPos = info.Value.initialPos;
                                        pp.linkRot = info.Value.initialRot.eulerAngles;
                                        cuaBones.Add(pp);
                                        cuaDic[info.Value.trns.name] = pp;
                                        logDebug($"extra bone {pp.target.name} binded");
                                    }
                                }
                                */
                                _currentBone = _currentBones[0];
                                RebuildMorphs();
                                RefreshUI();
                                logDebug("Seaching Rigidbody");
                                Rigidbody[] rbs = __parent.GetComponentsInChildren<Rigidbody>();
                                if (rbs.Length > 0) {
                                    foreach (Rigidbody rb in rbs) {
                                        if (cuaDic.ContainsKey(rb.name)) {
                                            cuaDic[rb.name].link = rb.transform;
                                            cuaDic[rb.name].rb = rb;
                                            logDebug($"Rigidbody found {rb.name}");
                                        }
                                    }
                                } else {
                                    logDebug("Rigidbody not found.");
                                }
                                break;
                            } else {
                                logDebug($"CUAAutoFollow : DAZCharacterSelector not Found on {__parent.name}");
                            }
                        } else {
                            logDebug($"CUAAutoFollow : {__parent.name} is not Person");
                        }
                    }
                    if (_characterSelector == null) {
                        logDebug("Missing DAZCharacterSelector");
                    }
                }
            }
        }

        void SyncWithBone() {
            if (_loading) return;
            if (!editorInitialized) return;
            Atom atom = GetContainingAtom();
            SkinnedMeshRenderer sm = _root.GetComponentInChildren<SkinnedMeshRenderer>();

            if (_characterSelector != null) {
                foreach (KeyValuePair<string, Dictionary<string, BoneOffset>> pair in morphDic) {
                    _rate[pair.Key] = 0;
                    foreach(MorphInfo info in _morphs[pair.Key]) {
                        _rate[pair.Key] += info.morph.morphValue;
                    }
                }

                foreach (KeyValuePair<string, float> p in _rate) {
                    if (_lastrate.ContainsKey(p.Key)) {
                        if (_lastrate[p.Key] != _rate[p.Key]) {
                            logDebug($"rate[{p.Key}] = {_rate[p.Key]}");
                            _dirty = true;
                        }
                        _lastrate[p.Key] = p.Value;
                    } else {
                        _dirty = true;
                        _lastrate[p.Key] = p.Value;
                    }

                }
                if (_dirty) {
                    logDebug("offset changing");
                    foreach (JSONStorableFloat jf in cuaMorphs) {
                        _blend[jf.name] = 0;
                    }
                    foreach (cuaBone b in cuaBones) {
                        Vector3 offsetpos = Vector3.zero;
                        Vector3 offsetrot = Vector3.zero;
                        Vector3 offsetscale = new Vector3(1.0f, 1.0f, 1.0f);
                        string key;
                        if (b.bone != null) {
                            key = b.bone.name;
                            //logDebug($"using bone name {b.bone.name}");
                        } else {
                            key = b.target.name;
                            //logDebug($"using target name {b.target.name}");
                        }
                        foreach (KeyValuePair<string, Dictionary<string, BoneOffset>> pair in morphDic) {
                            if (pair.Value.ContainsKey(key)) {
                                float rate = _rate[pair.Key];
                                offsetpos.x += pair.Value[key].offsetX * rate;
                                offsetpos.y += pair.Value[key].offsetY * rate;
                                offsetpos.z += pair.Value[key].offsetZ * rate;
                                offsetrot.x += pair.Value[key].rotateX * rate;
                                offsetrot.y += pair.Value[key].rotateY * rate;
                                offsetrot.z += pair.Value[key].rotateZ * rate;
                                offsetscale.x += (pair.Value[key].scaleX - 1.0f) * rate;
                                offsetscale.y += (pair.Value[key].scaleY - 1.0f) * rate;
                                offsetscale.z += (pair.Value[key].scaleZ - 1.0f) * rate;
                                foreach (KeyValuePair<string, float> m in pair.Value[key].morphs) {
                                    if (_blend.ContainsKey(m.Key)) {
                                        _blend[m.Key] += m.Value * rate;
                                    }
                                }
                            }
                        }
                        b.posOffset = offsetpos;
                        b.rotOffset = offsetrot;
                        b.scaleOffset = offsetscale;
                    }
                    foreach (cuaBone b in cuaBones) {
                        if (b.bone == null) {
                            b.target.localPosition = b.linkPos + b.posOffset;
                            b.target.localRotation = Quaternion.Euler(b.rotOffset.x, b.rotOffset.y, b.rotOffset.z);
                            b.target.localScale = b.scaleOffset;
                            //logDebug($"{b.target.name} changed with ({b.posOffset.x},{b.posOffset.y},{b.posOffset.z})");
                        }
                    }
                    _dirty = false;
                    logDebug("offset changed");
                }
                if (!posLinked) {
                    atom.mainController.transform.position = _parentAtom.transform.position;
                }
                if (!rotLinked) {
                    atom.mainController.transform.rotation = _parentAtom.transform.rotation;
                }

                //_root.localScale = new Vector3(jsScale.val,jsScale.val,jsScale.val);
                foreach (cuaBone b in cuaBones) {
                    if (b.bone == null) continue;
                    Quaternion q = Quaternion.Euler(b.rotOffset.x, b.rotOffset.y, b.rotOffset.z);
                    Vector3 offset = b.posOffset * jsScale.val;
                    offset = q * offset;

                    Vector3 pos;
                    Quaternion rot;
                    Vector3 scale;
                    if (b.link != null) {
                        pos = b.link.position + (b.link.localToWorldMatrix.MultiplyVector(offset));
                        rot = b.link.rotation * q;
                    } else {
                        pos = b.src.position + (b.src.localToWorldMatrix.MultiplyVector(offset));
                        rot = b.src.rotation * q;
                    }
                    if (b.posLinked) {
                        //pos = _root.position + b.link.localToWorldMatrix.MultiplyVector(offset);
                        pos = b.holdPos + (b.link.localToWorldMatrix.MultiplyVector(offset));
                    }
                    if (b.rotLinked) {
                        rot = b.holdRot * q;
                    }
                    scale = b.scaleOffset;
                    b.target.position = pos;
                    b.target.rotation = rot;
                    b.target.localScale = scale;
                }
                var mesh = sm.sharedMesh;
                foreach (KeyValuePair<string, float> pair in _blend) {
                    sm.SetBlendShapeWeight(mesh.GetBlendShapeIndex(pair.Key), pair.Value);
                }
            }
            ratecount++;
        }
        void LateUpdate() {
            if (!editorInitialized) return;
            SkinnedMeshRenderer sm = _root.GetComponentInChildren<SkinnedMeshRenderer>();
            Atom atom = GetContainingAtom();
            Rigidbody rbLink = atom.mainController.linkToRB;

            if (atom.mainController.currentPositionState == FreeControllerV3.PositionState.Hold) {
                if (!posLinked) {
                    logDebug("Calcurating LinkPos");
                    Vector3 initPos = _root.position;
                    foreach (cuaBone b in cuaBones) {
                        if (b.bone == null) continue;
                        if (b.link != null) {
                            if (rbLink.transform == b.link) {
                                _root.position = b.link.position;
                                b.linkPos = rbLink.transform.InverseTransformPoint(atom.childAtomContainer.position);
                                b.posLinked = true;
                                logDebug($"{b.src.name} linkPos updated");
                            } else {
                                b.posLinked = false;
                            }
                        }
                    }
                    _root.position = initPos;
                    posLinked = true;
                }
            } else {
                if (posLinked) {
                    foreach (cuaBone b in cuaBones) {
                        b.posLinked = false;
                    }
                }
                posLinked = false;
            }

            if (atom.mainController.currentRotationState == FreeControllerV3.RotationState.Hold && rbLink) {
                if (!rotLinked) {
                    Quaternion initRot = _root.rotation;
                    foreach (cuaBone b in cuaBones) {
                        if (b.bone == null) continue;
                        if (b.link != null) {
                            if (rbLink.transform == b.link) {
                                Transform src = b.link;
                                atom.mainController.transform.rotation = b.link.rotation;
                                _root.rotation = b.link.rotation;
                                b.linkRot = (Quaternion.Inverse(_root.rotation) * b.link.rotation).eulerAngles;
                                b.rotLinked = true;
                                logDebug("linkRot updated");
                            } else {
                                b.rotLinked = false;
                            }
                        }
                    }
                    _root.rotation = initRot;
                    rotLinked = true;
                }
            } else {
                if (rotLinked) {
                    foreach (cuaBone b in cuaBones) {
                        b.rotLinked = false;
                    }
                }
                rotLinked = false;
            }

            _root.localScale = new Vector3(jsScale.val, jsScale.val, jsScale.val);
            if (posLinked | rotLinked) {
                foreach (cuaBone b in cuaBones) {
                    if (b.posLinked | b.rotLinked) {
                        Vector3 pos = (b.posLinked ? b.link.transform.TransformPoint(b.linkPos) : atom.transform.position);
                        Quaternion rot = (b.rotLinked ? (b.link.transform.rotation * Quaternion.Inverse(Quaternion.Euler(b.linkRot))) : atom.transform.rotation);
                        atom.mainController.transform.SetPositionAndRotation(pos, rot);
                        b.holdPos = _root.position;
                        b.holdRot = _root.rotation;
                    }
                }
            }


            SyncWithBone();
        }

        private bool InitEditor() {
            if (editorInitialized == true) return true;
            _root = containingAtom.childAtomContainer;
            if (_root == null) {
                return false;
            }
            cuaBonedict.Clear();
            foreach (JSONStorableFloat jf in cuaMorphs) {
                RemoveSlider(jf);
            }
            cuaMorphs.Clear();
            SkinnedMeshRenderer sm = _root.GetComponentInChildren<SkinnedMeshRenderer>();
            if (sm != null) {
                var mesh = sm.sharedMesh;
                for (int i = 0; i < mesh.blendShapeCount; i++) {
                    JSONStorableFloat jf = new JSONStorableFloat(mesh.GetBlendShapeName(i), 0, OnChangeMorph, 0, 100f);
                    CreateSlider(jf, true);
                    cuaMorphs.Add(jf);
                }
                logDebug("bone Scan Started.");
                Transform[] bones = sm.rootBone.GetComponentsInChildren<Transform>();
                foreach (Transform bone in bones) {
                    cuaBoneInfo info = new cuaBoneInfo();
                    cuaBonedict[bone.name] = info;
                    cuaBonedict[bone.name].trns = bone;
                    cuaBonedict[bone.name].initialPos = bone.localPosition;
                    cuaBonedict[bone.name].initialRot = bone.localRotation;
                    logDebug($"{bone.name} pos = {cuaBonedict[bone.name].initialPos.x},{cuaBonedict[bone.name].initialPos.z},{cuaBonedict[bone.name].initialPos.z}");
                    logDebug($"{bone.name} rot = {cuaBonedict[bone.name].initialRot}");
                }
                logDebug("bone Scan End");
            } else {
                return false;
            }
            editorInitialized = true;
            return true;
        }

        private void logDebug(string debugText) {
            //SuperController.LogMessage("CUAAutoFollow : " + debugText);
        }

        private void logMessage(string debugText) {
            SuperController.LogMessage("CUAAutoFollow : " + debugText);
        }

        private void logError(string debugText) {
            SuperController.LogError("CUAAutoFollow : " + debugText);
        }
    }
}
